/*************************************************************************
 *
 *  OpenOffice.org - a multi-platform office productivity suite
 *
 *  $RCSfile: recordinfoprovider.hxx,v $
 *
 *  $Revision: 1.1.2.1 $
 *
 *  last change: $Author: dr $ $Date: 2007/11/20 15:56:41 $
 *
 *  The Contents of this file are made available subject to
 *  the terms of GNU Lesser General Public License Version 2.1.
 *
 *
 *    GNU Lesser General Public License Version 2.1
 *    =============================================
 *    Copyright 2005 by Sun Microsystems, Inc.
 *    901 San Antonio Road, Palo Alto, CA 94303, USA
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License version 2.1, as published by the Free Software Foundation.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA  02111-1307  USA
 *
 ************************************************************************/

#ifndef OOX_CORE_RECORDINFOPROVIDER_HXX
#define OOX_CORE_RECORDINFOPROVIDER_HXX

#include <map>
#include <boost/shared_ptr.hpp>
#include <sal/types.h>

namespace oox {
namespace core {

// ============================================================================

/** Describes record identifiers used to create contexts in a binary stream.

    If a record is used to start a new context, usually the record identifier
    increased by 1 is used to mark the end of this context, e.g. the Excel
    record SHEETDATA == 0x0091 starts the <sheetData> context, and the record
    SHEETDATA_END == 0x0092 ends this context. But some records are used to
    start a new context, though there is no identifier to end this context,
    e.g. the ROW, EXTROW, or EXTERNBOOK records. These record identifiers can
    be marked by setting the mnEndRecId member of this struct to -1.
 */
struct RecordInfo
{
    sal_Int32           mnStartRecId;       /// Record identifier for context start.
    sal_Int32           mnEndRecId;         /// Record identifier for context end, -1 = no record.
};

// ============================================================================

/** Provides a list of record identifiers used to start/end contexts in a
    binary record stream.

    This helper class is used by the RecordParser class to recognize contexts
    in a stream of binary records. Record identifiers without corresponding
    RecordInfo struct will be handled as 'simple records', means they are
    embedded in the current context and do not open a new context.
 */
class RecordInfoProvider
{
public:
    /** Constructs a new provider from the passed list of RecordInfo structs.

        @param pRecs  The list of RecordInfo structs. The list must be
            terminated by an element containing -1 as start identifier.
     */
    explicit            RecordInfoProvider( const RecordInfo* pRecs );

    virtual             ~RecordInfoProvider();

    /** Returns a RecordInfo struct that contains the passed record identifier
        as context start identifier.
     */
    const RecordInfo*   getStartRecordInfo( sal_Int32 nRecId ) const;

    /** Returns a RecordInfo struct that contains the passed record identifier
        as context end identifier.
     */
    const RecordInfo*   getEndRecordInfo( sal_Int32 nRecId ) const;

private:
    typedef ::std::map< sal_Int32, RecordInfo > RecordInfoMap;
    RecordInfoMap       maStartMap;
    RecordInfoMap       maEndMap;
};

typedef ::boost::shared_ptr< RecordInfoProvider > RecordInfoProviderRef;

// ============================================================================

} // namespace core
} // namespace oox

#endif

